﻿/*
 * Peony-Qt
 *
 * Copyright (C) 2024, Tianjin KYLIN Information Technology Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 * @author: Renyg <renyangguang@kylinos.cn>
 * @date:   2024-08-02 17:34
 */

#ifndef _KYLIN_SINGLETON_UTILS_H_
#define _KYLIN_SINGLETON_UTILS_H_

#include <string>
#include <functional>
#include <map>
#include <stack>
#include <mutex>
#include <vector>
#include "macro-utils.h"
#include "peony-core_global.h"

//using namespace Peony;

//namespace Peony {

typedef std::function<void()> DelSingletonAction;
typedef std::stack<DelSingletonAction> SingletonDeleteStack;

void PEONYCORESHARED_EXPORT pushSingletonAction(DelSingletonAction delAction);
void PEONYCORESHARED_EXPORT destroySingleton();

extern std::map<std::string, void*> PEONYCORESHARED_EXPORT g_singletons;
extern std::recursive_mutex	PEONYCORESHARED_EXPORT g_singletons_mutex;

template<class T>
class PEONYCORESHARED_EXPORT KylinSingleton
{
public:
	template<typename... Args>
	static T*	instance(Args&&... args)
	{
		static T* _instance = nullptr;
		static std::once_flag oc;
		std::call_once(oc, [&] {
			_instance = new T(std::forward<Args>(args)...);
			std::lock_guard<std::recursive_mutex> lg(g_singletons_mutex);
            g_singletons[typeid(T).name()] = _instance;
			if (_instance->AutoDestroy)
			{
				pushSingletonAction([=] { 
					qDebug("Delete Singleton : %s", typeid(T).name()); 
                    g_singletons[typeid(T).name()] = nullptr;
					delete _instance; 
					_instance = nullptr;
				});
			}
		});
		return _instance;
	}

	static bool isDestroyed()
	{
		std::string clsName = typeid(T).name();
		std::lock_guard<std::recursive_mutex> lg(g_singletons_mutex);
		return g_singletons.find(clsName) != g_singletons.end() && g_singletons[clsName] == nullptr;
	}

private:
    KylinSingleton() {}
    virtual ~KylinSingleton() {}
    KylinSingleton(const KylinSingleton& st) {}
    KylinSingleton &operator=(const KylinSingleton& st) {}
};

typedef std::function<void()> SingletonInit;
typedef std::vector<SingletonInit>	SingletonInitList;
typedef Instance<SingletonInitList>	SingletonInitializer;

#define DECLARESINGLETON(clsname) public: friend class KylinSingleton<clsname>; bool AutoDestroy = true;
#define DECLARESINGLETON_NOTAUTODESTROY(clsname) public: friend class KylinSingleton<clsname>; bool AutoDestroy = false;

#define SINGLETON(x,...) KylinSingleton<x>::instance(__VA_ARGS__)
#define DESTROYSINGLETON	{ destroySingleton(); }
#define SINGLETONDESTROYED(T) KylinSingleton<T>::isDestroyed()
#define PREINITSINGLETON(T) Global(T##_PreInit_){ \
SingletonInitializer::instance().push_back([]{SINGLETON(T);});\
}

//}
#endif  // _KYLIN_SINGLETON_UTILS_H_


